(* some common operations on bdds *)

(* John D. Ramsdell -- Decemeber 1998 *)

(* Copyright (C) 2001 The MITRE Corporation

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA. *)

open Bdd

type path = (bool * int) list (* See interface for description *)

let conjoin x y = ite(x, y, falsehood)

let disjoin x y = ite(x, truth, y)

let difference x y = conjoin x (complement y)

let implies x y = ite(x, y, truth)

let restrict value i bdd =
  if not (is_node bdd) then
    bdd
  else
    let v = variable bdd in
    if i != v then
      bdd
    else
      if value then
	consequent bdd
      else
	alternative bdd

let prod offset size f g =
  let rec recur f g =
    if is_falsehood f || is_falsehood g then
      falsehood
    else if is_truth f && is_truth g then
      truth
    else
      let x = variable f in
      let y = variable g in
      let z = min x y in
      let r0 = recur (restrict false z f) (restrict false z g) in
      let r1 = recur (restrict true z f) (restrict true z g) in
      if z >= offset && z < offset + size then
	disjoin r0 r1
      else
	ite(identity z, r1, r0) in
  recur f g

let increase_variable_names offset size bdd =
  let rec recur bdd =
    if is_node bdd then
      let v = variable bdd in
      let c = recur (consequent bdd) in
      let a = recur (alternative bdd) in
      let v' =
	if v < offset then
	  v
	else
	  v + size in
      ite(identity v', c, a)
    else
      bdd in
  recur bdd

let decrease_variable_names offset size bdd =
  let rec recur bdd =
    if is_node bdd then
      let v = variable bdd in
      let c = recur (consequent bdd) in
      let a = recur (alternative bdd) in
      if v >= offset - size && v < offset then
	disjoin c a
      else
	let v' =
	  if v >= offset then
	    v - size
	  else
	    v in
	ite(identity v', c, a)
    else
      bdd in
  recur bdd

let rename offset size bdd =
  if size > 0 then
    increase_variable_names offset size bdd
  else if size < 0 then
    decrease_variable_names offset (-size) bdd
  else
    bdd

let collapse offset size bdd =
  let rec recur bdd =
    if is_node bdd then
      let v = variable bdd in
      let c = recur (consequent bdd) in
      let a = recur (alternative bdd) in
      if v >= offset && v < offset + size then
	ite (identity v, c, a)
      else
	disjoin c a
    else
      bdd in
  recur bdd

let path2bdd path =
  let rec loop path bdd =
    match path with
      [] -> bdd
    | (negate, var) :: p ->
	let b = identity var in
	loop p (conjoin (if negate then complement b else b) bdd) in
  loop path truth
