open Hashtbl
open Bdd
open Bdd_utils
open Bits
open Formulas

type descriptor = {
    clss_size : int;
    perms_size : int;
    types_size : int;
    roles_size : int;
    users_size : int;
    clss_offset : int;
    perms_offset : int;
    in_okay_offset : int;
    in_types_offset : int;
    in_roles_offset : int;
    in_users_offset : int;
    out_okay_offset : int;
    out_types_offset : int;
    out_roles_offset : int;
    out_users_offset : int;
    state_size : int;
    in_state_offset: int;
    out_state_offset: int;
    (* The lookup functions raise Not_found when a mapping for
       a name is not available. *)
    lookup_cls : string -> bdd;
    lookup_perm : string -> bdd;
    lookup_in_okay : bool -> bdd;
    lookup_in_type : string -> bdd;
    lookup_in_role : string -> bdd;
    lookup_in_user : string -> bdd;
    lookup_out_okay : bool -> bdd;
    lookup_out_type : string -> bdd;
    lookup_out_role : string -> bdd;
    lookup_out_user : string -> bdd }

(* Lookup names in a hash table *)

let lookup = Hashtbl.find

(* hash tables map list elements to a bdd *)

let list2table offset size list =
  let tbl = create size in
  let rec loop i set =
    match set with
      [] -> tbl
    | elem :: set ->
	let bdd = path2bdd (int2path offset size i) in
	add tbl elem bdd;
	loop (i + 1) set in
  loop 0 list

let lookup_bool i b =
  let bdd = identity i in
  match b with
    true -> bdd
  | false -> complement bdd

let make_descriptor
    { types = types;
      roles = roles;
      users = users;
      classes = classes;
      permissions = permissions
    } =
  let okay_size = 1 in
  let clss_size = bit_length (List.length classes) in
  let perms_size = bit_length (List.length permissions) in
  let types_size = bit_length (List.length types) in
  let roles_size = bit_length (List.length roles) in
  let users_size = bit_length (List.length users) in

  let state_size = okay_size + types_size + roles_size + users_size in
  let total_size = clss_size + perms_size + 2 * state_size in
  let in_state_offset = perms_size + clss_size in
  let out_state_offset = in_state_offset + state_size in

  set_var_num total_size;

  let perms_offset = 0 in
  let clss_offset = perms_offset + perms_size in
  let in_okay_offset = in_state_offset in
  let in_users_offset = in_okay_offset + okay_size in
  let in_roles_offset = in_users_offset + users_size in
  let in_types_offset = in_roles_offset + roles_size in

  let out_okay_offset = in_okay_offset + state_size in
  let out_types_offset = in_types_offset + state_size in
  let out_roles_offset = in_roles_offset + state_size in
  let out_users_offset = in_users_offset + state_size in

  let clss_tbl = list2table clss_offset clss_size classes in
  let perms_tbl = list2table perms_offset perms_size permissions in
  let in_types_tbl =
    list2table in_types_offset types_size types in
  let in_roles_tbl = list2table in_roles_offset roles_size roles in
  let in_users_tbl = list2table in_users_offset users_size users in
  let out_types_tbl =
    list2table out_types_offset types_size types in
  let out_roles_tbl = list2table out_roles_offset roles_size roles in
  let out_users_tbl = list2table out_users_offset users_size users in

  { clss_size = clss_size;
    perms_size = perms_size;
    types_size = types_size;
    roles_size = roles_size;
    users_size = users_size;
    clss_offset = clss_offset;
    perms_offset = perms_offset;
    in_okay_offset = in_okay_offset;
    in_types_offset = in_types_offset;
    in_roles_offset = in_roles_offset;
    in_users_offset = in_users_offset;
    out_okay_offset = out_okay_offset;
    out_types_offset = out_types_offset;
    out_roles_offset = out_roles_offset;
    out_users_offset = out_users_offset;
    state_size = state_size;
    in_state_offset = in_state_offset;
    out_state_offset = out_state_offset;
    lookup_cls = lookup clss_tbl;
    lookup_perm = lookup perms_tbl;
    lookup_in_okay = lookup_bool in_okay_offset;
    lookup_in_type = lookup in_types_tbl;
    lookup_in_role = lookup in_roles_tbl;
    lookup_in_user = lookup in_users_tbl;
    lookup_out_okay = lookup_bool out_okay_offset;
    lookup_out_type = lookup out_types_tbl;
    lookup_out_role = lookup out_roles_tbl;
    lookup_out_user = lookup out_users_tbl
  }
