open Formulas

(* NuSMV formulas *)

type var =
    Var of string
  | Nxt of string

type form =
    Form_true
  | Form_false
  | Form_in of var * string list
  | Form_eq of var * var
  | Form_not of form
  | Form_and of form * form
  | Form_or of form * form
  | Form_imply of form * form
  | Form_iff of form * form
	(* CTL *)
  | Form_exists_next_state of form
  | Form_forall_next_state of form
  | Form_exists_finally of form
  | Form_forall_finally of form
  | Form_exists_globally of form
  | Form_forall_globally of form
  | Form_exists_until of form * form
  | Form_forall_until of form * form
	(* LTL *)
  | Form_next_state of form
  | Form_finally of form
  | Form_globally of form
  | Form_until of form * form

open Unparse

(* Unparser operators *)

let form_true = Atom "TRUE"
let form_false = Atom "FALSE"
let form_atom str = Atom str

let comma_op =
  { prec = 0;
    assoc = Assoc;
    printer = fun () ->
      Format.print_string ",";
      Format.print_space()
  }

let form_comma l r =
  Binary (comma_op, l, r)

let imply_op =
  { prec = 1;
    assoc = Right;
    printer = fun () ->
      Format.print_space();
      Format.print_string "-> "
  }

let form_imply l r =
  Binary (imply_op, l, r)

let iff_op =
  { prec = 2;
    assoc = Left;
    printer = fun () ->
      Format.print_space();
      Format.print_string "<-> "
  }

let form_iff l r =
  Binary (iff_op, l, r)

let or_op =
  { prec = 3;
    assoc = Assoc;
    printer = fun () ->
      Format.print_space();
      Format.print_string "| ";
  }

let form_or l r =
  Binary (or_op, l, r)

let and_op =
  { prec = 4;
    assoc = Assoc;
    printer = fun () ->
      Format.print_space();
      Format.print_string "& "
  }

let form_and l r =
  Binary (and_op, l, r)

(* Note the low precedence of not.  This is how NuSMV does it. *)
let not_op =
  { prec = 5;
    assoc = Right;
    printer = fun () ->
      Format.print_string "!"
  }

let form_not r =
  Unary (not_op, r)

let temp_op prec name =
  { prec = prec;
    assoc = Right;
    printer = fun () ->
      Format.print_string name;
      Format.print_string " "
  }

(* CTL *)

let ex_op = temp_op 6 "EX"
let form_ex r = Unary(ex_op, r)
let ax_op = temp_op 6 "AX"
let form_ax r = Unary(ax_op, r)
let ef_op = temp_op 6 "EF"
let form_ef r = Unary(ef_op, r)
let af_op = temp_op 6 "AF"
let form_af r = Unary(af_op, r)
let eg_op = temp_op 6 "EG"
let form_eg r = Unary(eg_op, r)
let ag_op = temp_op 6 "AG"
let form_ag r = Unary(ag_op, r)

let ctl_until_op =
  { prec = 0;
    assoc = Non_assoc;
    printer = fun () ->
      Format.print_space();
      Format.print_string "U "
  }

let form_eu l r =
  Match ("E [ ", Binary(ctl_until_op, l, Match ("", r, "")), " ]")

let form_au l r =
  Match ("A [ ", Binary(ctl_until_op, l, Match ("", r, "")), " ]")

(* LTL *)

let ltl_until_op =
  { prec = 6;
    assoc = Non_assoc;
    printer = fun () ->
      Format.print_space();
      Format.print_string "U "
  }

let form_u l r = Binary(ltl_until_op, l, r)

let x_op = temp_op 7 "X"
let form_x r = Unary(x_op, r)
let f_op = temp_op 7 "F"
let form_f r = Unary(f_op, r)
let g_op = temp_op 7 "G"
let form_g r = Unary(g_op, r)

let eq_op =
  { prec = 8;
    assoc = Non_assoc;
    printer = fun () ->
      Format.print_string " = "
  }

let form_eq l r =
  Binary(eq_op, l, r)

let neq_op =
  { prec = 8;
    assoc = Non_assoc;
    printer = fun () ->
      Format.print_string " != "
  }

let form_not_eq l r =
  Binary(neq_op, l, r)

let in_op =
  { prec = 9;
    assoc = Non_assoc;
    printer = fun () ->
      Format.print_string " in "
  }

let is_length_one l =
  match l with
    [] -> false
  | _ :: l' -> l' = []

let form_set s =
  let rec loop u s =
    match s with
      [] -> u
    | e :: s ->
	loop (form_comma u (Atom e)) s in
  match s with
    [] -> Atom "{}"
  | e :: s ->
      Match ("{", loop (Atom e) s, "}")

let form_in l r =
  if r = [] then
    form_false
  else if is_length_one r then
    form_eq l (form_atom (List.hd r))
  else
    Binary (in_op, l, form_set r)

let form_not_in l r =
  if r = [] then
    form_true
  else if is_length_one r then
    form_not_eq l (form_atom (List.hd r))
  else
    form_not (Binary (in_op, l, form_set r))

let unparse_var var =
  match var with
    Var name -> form_atom name
  | Nxt name -> form_atom ("next(" ^ name ^ ")")

let rec unparse_form form =
  match form with
    Form_true -> form_true
  | Form_false -> form_false
  | Form_not (Form_in (var, set)) ->
      form_not_in (unparse_var var) set
  | Form_in (var, set) ->
      form_in (unparse_var var) set
  | Form_not (Form_eq (var, var')) ->
      form_not_eq (unparse_var var) (unparse_var var')
  | Form_eq (var, var') ->
      form_eq (unparse_var var) (unparse_var var')
  | Form_not arg -> form_not (unparse_form arg)
  | Form_and (arg1, arg2) ->
      form_and (unparse_form arg1) (unparse_form arg2)
  | Form_or (arg1, arg2) ->
      form_or (unparse_form arg1) (unparse_form arg2)
  | Form_imply (arg1, arg2) ->
      form_imply (unparse_form arg1) (unparse_form arg2)
  | Form_iff (arg1, arg2) ->
      form_iff (unparse_form arg1) (unparse_form arg2)
  | Form_exists_next_state arg -> form_ex (unparse_form arg)
  | Form_forall_next_state arg -> form_ax (unparse_form arg)
  | Form_exists_finally arg -> form_ef (unparse_form arg)
  | Form_forall_finally arg -> form_af (unparse_form arg)
  | Form_exists_globally arg -> form_eg (unparse_form arg)
  | Form_forall_globally arg -> form_ag (unparse_form arg)
  | Form_exists_until (arg1, arg2) ->
      form_eu (unparse_form arg1) (unparse_form arg2)
  | Form_forall_until (arg1, arg2) ->
      form_au (unparse_form arg1) (unparse_form arg2)
  | Form_next_state arg -> form_x (unparse_form arg)
  | Form_finally arg -> form_f (unparse_form arg)
  | Form_globally arg -> form_g (unparse_form arg)
  | Form_until (arg1, arg2) ->
      form_u (unparse_form arg1) (unparse_form arg2)

let print_form form =
  unparse (unparse_form form)

let class_var_name = Var "c"
let permission_var_name = Var "p"
let type_var_name = Var "t"
let role_var_name = Var "r"
let user_var_name = Var "u"
let okay_var_name = Var "k"
let next_type_var_name = Nxt "t"
let next_role_var_name = Nxt "r"
let next_user_var_name = Nxt "u"
let next_okay_var_name = Nxt "k"

let has_suffix s x =
  let s_n = String.length s in
  let x_n = String.length x in
  if s_n < x_n then
    false
  else
    x = String.sub s (s_n - x_n) x_n

(* Annotate names for output *)
let anno_cls name = name ^ "_c"
let anno_perm name = name ^ "_p"
let anno_sid name = name ^ "_s"
let anno_attr name = name ^ "_a"
let anno_type name =
  if has_suffix name "_t" then
    name
  else
    failwith ("bad type name: " ^ name)
let anno_role name =
  if has_suffix name "_r" then
    name
  else
    failwith ("bad role name: " ^ name)
let anno_user name =
  if has_suffix name "_u" then
    name
  else
    name ^ "_u"				(* This is dangerous! *)

(* Uncomment to check output renaming. *)
(*
let anno_type name = name ^ "_t"
let anno_role name = name ^ "_r"
let anno_user name = name ^ "_u"
*)

(* Annotate sets for output *)
let anno_clss set = List.map anno_cls set
let anno_perms set = List.map anno_perm set
let anno_types set = List.map anno_type set
let anno_roles set = List.map anno_role set
let anno_users set = List.map anno_user set

let mk_true = Form_true
let mk_false = Form_false

let mk_not r =
  match r with
    Form_true -> Form_false
  | Form_false -> Form_true
  | Form_not f -> f
  | _ -> Form_not r

let mk_and l r =
  match (l, r) with
    (Form_true, f) -> f
  | (Form_false, _) -> Form_false
  | (f, Form_true) -> f
  | (_, Form_false) -> Form_false
  | _ -> Form_and (l, r)
	
let mk_or l r =
  match (l, r) with
    (Form_false, f) -> f
  | (Form_true, _) -> Form_true
  | (f, Form_false) -> f
  | (_, Form_true) -> Form_true
  | _ -> Form_or (l, r)
	
let mk_imply l r =
  match (l, r) with
    (Form_false, _) -> Form_true
  | (Form_true, f) -> f
  | (f, Form_false) -> mk_not f
  | (_, Form_true) -> Form_true
  | _ -> Form_imply (l, r)

let mk_iff l r =
  match (l, r) with
    (Form_false, f) -> mk_not f
  | (Form_true, f) -> f
  | (f, Form_false) -> mk_not f
  | (f, Form_true) -> f
  | _ -> Form_iff (l, r)

let mk_eq l r =
  Form_eq (l, r)

let mk_in l r =
  Form_in (l, r)

(* These operators could float a negation up through them. *)
(* For example, mk_ex (Form_not f) = mk_not (mk_ax f) *)
let mk_ex r = Form_exists_next_state r
let mk_ax r = Form_forall_next_state r
let mk_ef r = Form_exists_finally r
let mk_af r = Form_forall_finally r
let mk_eg r = Form_exists_globally r
let mk_ag r = Form_forall_globally r
let mk_x r = Form_next_state r
let mk_f r = Form_finally r
let mk_g r = Form_globally r

let mk_eu l r =
  match l with
    Form_true -> mk_ef r
  | _ -> Form_exists_until (l, r)

let mk_u l r =
  match l with
    Form_true -> mk_f r
  | _ -> Form_until (l, r)

let mk_assert v =
  mk_in v ["TRUE"]

let rec ctx2form ctx =
  match ctx with
    Ctx_true -> mk_true
  | Ctx_false -> mk_false
  | Ctx_types set -> mk_in type_var_name (anno_types set)
  | Ctx_roles set -> mk_in role_var_name (anno_roles set)
  | Ctx_users set -> mk_in user_var_name (anno_users set)
  | Ctx_not arg -> mk_not (ctx2form arg)
  | Ctx_and (arg1, arg2) ->
      mk_and (ctx2form arg1) (ctx2form arg2)
  | Ctx_or (arg1, arg2) ->
      mk_or (ctx2form arg1) (ctx2form arg2)
  | Ctx_imply (arg1, arg2) ->
      mk_imply (ctx2form arg1) (ctx2form arg2)
  | Ctx_iff (arg1, arg2) ->
      mk_iff (ctx2form arg1) (ctx2form arg2)

let print_init ctx =
  let ctx =
    mk_imply
      (mk_assert okay_var_name)
      (ctx2form ctx) in
  print_form ctx

let rec act2form act =
  match act with
    Act_true -> mk_true
  | Act_false -> mk_false
  | Act_classes set -> mk_in class_var_name (anno_clss set)
  | Act_permissions set -> mk_in permission_var_name (anno_perms set)
  | Act_not arg -> mk_not (act2form arg)
  | Act_and (arg1, arg2) ->
      mk_and (act2form arg1) (act2form arg2)
  | Act_or (arg1, arg2) ->
      mk_or (act2form arg1) (act2form arg2)
  | Act_imply (arg1, arg2) ->
      mk_imply (act2form arg1) (act2form arg2)
  | Act_iff (arg1, arg2) ->
      mk_iff (act2form arg1) (act2form arg2)

let rec tran2form tran =
  match tran with
    Tran_true -> mk_true
  | Tran_false -> mk_false
  | Tran_classes set -> mk_in class_var_name (anno_clss set)
  | Tran_permissions set -> mk_in permission_var_name (anno_perms set)
  | Tran_types set -> mk_in type_var_name (anno_types set)
  | Next_types set -> mk_in next_type_var_name (anno_types set)
  | Same_types -> mk_eq type_var_name next_type_var_name
  | Tran_roles set -> mk_in role_var_name (anno_roles set)
  | Next_roles set -> mk_in next_role_var_name (anno_roles set)
  | Same_roles -> mk_eq role_var_name next_role_var_name
  | Tran_users set -> mk_in user_var_name (anno_users set)
  | Next_users set -> mk_in next_user_var_name (anno_users set)
  | Same_users -> mk_eq user_var_name next_user_var_name
  | Tran_not arg -> mk_not (tran2form arg)
  | Tran_and (arg1, arg2) ->
      mk_and (tran2form arg1) (tran2form arg2)
  | Tran_or (arg1, arg2) ->
      mk_or (tran2form arg1) (tran2form arg2)
  | Tran_imply (arg1, arg2) ->
      mk_imply (tran2form arg1) (tran2form arg2)
  | Tran_iff (arg1, arg2) ->
      mk_iff (tran2form arg1) (tran2form arg2)

let print_trans tran =
  let tran =
    mk_iff
      (mk_assert next_okay_var_name)
      (mk_and
	 (mk_assert okay_var_name)
	 (tran2form tran)) in
  print_form tran

let emit_section title =
  Format.print_newline();
  Format.print_string "-- ";
  Format.print_string title;
  Format.print_newline()

let print_section_start name =
  Format.print_newline();
  Format.print_string " ";
  Format.print_string name;
  Format.print_newline();
  Format.print_newline();
  Format.print_string "  "

let print_var var =
  match var with
    Var name -> Format.print_string name
  | _ -> failwith "var def cannot be next"

let emit_state_var var_name var_type =
  Format.open_box 2;
  Format.print_string " VAR ";
  print_var var_name;
  Format.print_string ":";
  Format.print_space();
  unparse (form_set var_type);
  Format.print_string ";";
  Format.close_box();
  Format.print_newline()

let emit_bool_var var_name =
  Format.print_string " VAR ";
  print_var var_name;
  Format.print_string ": boolean;";
  Format.print_newline()

let emit_decls
    {types = types;
     roles = roles;
     users = users;
     classes = classes;
     permissions = permissions} =
  Format.print_string "MODULE main";
  Format.print_newline();
  Format.print_newline();
  emit_state_var type_var_name (anno_types types);
  emit_state_var role_var_name (anno_roles roles);
  emit_state_var user_var_name (anno_users users);
  emit_state_var class_var_name (anno_clss classes);
  emit_state_var permission_var_name (anno_perms permissions);
  emit_bool_var okay_var_name

let emit_init contexts =
  print_section_start "INIT";
  print_init contexts;
  Format.print_newline()

let emit_trans transitions =
  print_section_start "TRANS";
  print_trans transitions;
  Format.print_newline()

let rec unnext tran =
  match tran with
    Tran_true -> tran
  | Tran_false -> tran
  | Tran_classes set -> tran
  | Tran_permissions set -> tran
  | Next_types set -> Tran_types set
  | Next_roles set -> Tran_roles set
  | Next_users set -> Tran_users set
  | Same_types -> tran
  | Same_roles -> tran
  | Same_users -> tran
  | Tran_not arg -> Tran_not (unnext arg)
  | Tran_and (arg1, arg2) ->
      Tran_and (unnext arg1, unnext arg2)
  | Tran_or (arg1, arg2) ->
      Tran_or (unnext arg1, unnext arg2)
  | Tran_imply (arg1, arg2) ->
      Tran_imply (unnext arg1, unnext arg2)
  | Tran_iff (arg1, arg2) ->
      Tran_iff (unnext arg1, unnext arg2)
  | _ -> failwith "bad spec form: not a next state"

let print_spec mk_next_state spec =
  match spec with
    Tran_imply (hypth, concl) ->
      let hypth = tran2form hypth in
      let concl = tran2form (unnext concl) in
      let u =
	mk_imply
	  hypth
	  (mk_next_state (mk_imply (mk_assert okay_var_name) concl)) in
      print_form u
  | _ -> failwith "bad spec form: not an implication"

let print_spec_section_start ltl =
  if ltl then
    print_section_start "LTLSPEC"
  else
    print_section_start "SPEC"
	
let emit_spec ltl transitions =
  print_spec_section_start ltl;
  if ltl then
    print_spec mk_x transitions
  else
    print_spec mk_ax transitions;
  Format.print_newline()

(* When ltl is true, emit LTL specs, otherwise emit CTL. *)
let nusmv ltl lts =
  emit_decls lts;
  emit_section "Initial States";
  emit_init lts.initial;
  emit_section "Transition Relation";
  emit_trans lts.transition;
  let specs = lts.specifications in
  if specs <> [] then
    begin
      emit_section "Specifications";
      List.iter (emit_spec ltl) specs
    end

let mk_okay_contexts contexts =
  mk_and (mk_assert okay_var_name) contexts

let mk_next ltl r =
  if ltl then
    mk_x r
  else
    mk_ex r

let mk_until ltl l r =
  if ltl then
    mk_u l r
  else
    mk_eu l r

let mk_next_or_until ltl more l r =
  let z =
    if more then
      mk_until ltl l r
    else
      r in
  mk_and l (mk_next ltl z)

let emit_follows_diagram ltl exceptions first label more arrows last =
  let rest = mk_until ltl (mk_not exceptions) (mk_okay_contexts last) in
  let mk_rest state label more next =
    if not more then
      mk_and (mk_not next) rest
    else
      mk_until
	ltl
	(mk_not (mk_or next exceptions))
	(mk_and
	   (mk_not
	      (mk_or
		 label
		 next))
	   rest) in
  let mk_base state label more next =
    mk_and
      state
      (mk_or
	 (mk_and (mk_not label) rest)
	 (mk_and
	    (mk_and label (mk_not exceptions))
	    (mk_next ltl (mk_rest state label more next)))) in
  let rec recur i j state label more arrows =
    match arrows with
      [] -> mk_base state label more last
    | (next, new_label, new_more) :: arrows ->
	if j < i then
	  mk_and
	    state
	    (mk_next_or_until
	       ltl
	       more
	       (mk_and label (mk_not exceptions))
	       (recur i (j + 1) next new_label new_more arrows))
	else
	  mk_base state label more next in
  let n = List.length arrows in
  for i = 0 to n do
    print_spec_section_start ltl;
    let u = recur i 0 first label more arrows in
    print_form (mk_not u);
    Format.print_newline()
  done

let emit_ordered ltl exceptions first state next last =
  print_spec_section_start ltl;
  let u =
    mk_and
      first
      (mk_until
	 ltl
	 (mk_not (mk_or state exceptions))
	 (mk_and
	    next
	    (mk_until ltl (mk_not exceptions) (mk_okay_contexts last)))) in
  print_form (mk_not u);
  Format.print_newline()

let emit_ordered_diagram ltl exceptions first arrows last =
  let rec loop state arrows =
    match arrows with
      [] -> emit_ordered ltl exceptions first state last last
    | (next, _, _) :: arrows ->
	emit_ordered ltl exceptions first state next last;
	loop next arrows in
  match arrows with
    [] -> ()
  | (state, _, _) :: arrows ->
      loop state arrows

(* When ltl is true, emit LTL specs, otherwise emit CTL. *)
let emit_diagram ltl (arrows, last, ex_ctx, ex_act) =
  let exceptions =
    mk_or (ctx2form ex_ctx) (act2form ex_act) in
  let mk_states (s, l, b) =
    ctx2form s, act2form l, b in
  let arrows = List.map mk_states arrows in
  let last = ctx2form last in
  match arrows with
      [] -> ()
    | (first, label, more) :: arrows ->
	emit_section "Diagram";
	emit_follows_diagram ltl exceptions first label more arrows last;
	emit_ordered_diagram ltl exceptions first arrows last
