(* This section is all for argument processing *)

(* Bdd defaults *)

let utn = ref Bdd.init_defaults.Bdd.utn
let uts = ref Bdd.init_defaults.Bdd.uts
let ctn = ref Bdd.init_defaults.Bdd.ctn
let cts = ref Bdd.init_defaults.Bdd.cts

exception Bad_arg of int

let check_arg loc int =
  if int > 0 then
    loc := int
  else
    raise (Bad_arg int)

let default loc =
  " (default " ^ string_of_int !loc ^ ")"

(* Mode *)

let version = ref false

(* file names *)

let output_file = ref "-"

let lts_file = ref "-"

let tran_file = ref "-"

let anon_var = ref 0

let anon_fun str =
  match !anon_var with
    0 ->
      lts_file := str;
      anon_var := 1
  | 1 ->
      tran_file := str;
      anon_var := 2
  | _ ->
      anon_var := 3

(* set up Arg.parse *)

let arg_spec =
  [("-o", Arg.String(fun s -> output_file := s),
    "name  - send output to this file (default stdout)" );
   ("-v", Arg.Set(version),
    "      - display version information and exit");
   ("-utn", Arg.Int(check_arg utn),
    "int - unique table number" ^ default utn);
   ("-uts", Arg.Int(check_arg uts),
    "int - unique table size" ^ default uts);
   ("-ctn", Arg.Int(check_arg ctn),
    "int - computed table number" ^ default ctn);
   ("-cts", Arg.Int(check_arg cts),
    "int - computed table size" ^ default cts);
   ("--", Arg.Rest(anon_fun),
    "      - treat remaining args as file names, where - means stdin")]

let usage_msg = "Usage: typesallowed lts [tran]";;

Arg.parse arg_spec anon_fun usage_msg;

if !version then
  begin
    Print_version.print_version();
    exit 0
  end;

if !anon_var < 1 || !anon_var > 2 then
  begin
    prerr_string "bad arg count";
    prerr_newline();
    prerr_string usage_msg;
    prerr_newline();
    exit 1
  end;

(* init bdd *)

Bdd.init(!utn, !uts, !ctn, !cts);;

(* file name handling *)

let lts_file_name = !lts_file in
let tran_file_name = !tran_file in
let output_file_name = !output_file in

if lts_file_name = tran_file_name then
  begin
    prerr_string "input files must differ";
    prerr_newline();
    prerr_string usage_msg;
    prerr_newline();
    exit 1
  end;

(* read input *)

try

  let lts = Read_formulas.read_lts lts_file_name in

  let tran = Read_formulas.read_tran tran_file_name in

  let desc = Descriptor.make_descriptor lts in

  let ans_bdd =
    match lts with
      { Formulas.types = types;
	Formulas.roles = roles;
	Formulas.users = users;
	Formulas.transition = transition
      } ->
	let ans_tran = Formulas.Tran_and(transition, tran) in
	let bdd = Tran2bdd.tran2bdd desc types roles users ans_tran in
	let offset = desc.Descriptor.out_types_offset in
	let size = desc.Descriptor.types_size in
	Bdd_utils.collapse offset size bdd in

  let probe_name lookup name =
    let bdd = lookup name in
    not (Bdd.is_falsehood (Bdd_utils.conjoin bdd ans_bdd)) in

  let next_types =
    let pred = probe_name desc.Descriptor.lookup_out_type in
    List.filter pred lts.Formulas.types in

(* open output file *)

  let out =
    if output_file_name = "-" then
      stdout
    else
      open_out output_file_name in

  Format.set_formatter_out_channel out;

(* produce output *)

  Formulas.print_transitions (Formulas.Next_types next_types);
  Format.print_newline();
  exit 0

with Failure s ->
  prerr_string s;
  prerr_newline();
  exit 1
